{{ sdram-video-40MHz-le-propfont.spin }}
CON
  Q_SIZE = 300
  Q_BYTES = Q_SIZE << 2

  WAITCYCLES = 34
  WIDTH = 800
  HEIGHT = 480
  HGAP = 100      ' pixels per hblank and vblank
  VGAP = 100


  pin_BA0  = 13
  pin_BA1  = 14
  pin_SCLK = 15

  pin_DQM  = 16  ' waitvid controlled
  pin_LE   = 17
  pin_WE   = 18
  pin_CAS  = 19
  pin_RAS  = 20
  pin_CS   = 21
  pin_CKE  = 22  ' waitvid controlled
  pin_DEN  = 23

  pin_DCLK = 24
  pin_LCLK = 25

  DENPIN  = %10000000   ' these are video byte masks
  CEPIN   = %01000000
  CSPIN   = %00100000
  RASPIN  = %00010000
  CASPIN  = %00001000
  WEPIN   = %00000100
  LEPIN	  = %00000010
  DQMPIN  = %00000001


  IDLE       =	CSPIN | RASPIN | CASPIN | WEPIN | LEPIN
  NONE	     =	CEPIN | RASPIN | CASPIN | WEPIN | LEPIN
  DEN        =	DENPIN | CEPIN | RASPIN | CASPIN | WEPIN | LEPIN
  LE         =	CEPIN | RASPIN | CASPIN | WEPIN 
  LE_DQM     =	CEPIN | RASPIN | CASPIN | WEPIN | DQMPIN
  DQM        =	CEPIN | RASPIN | CASPIN | WEPIN | LEPIN | DQMPIN
  ACTIVE     =	CEPIN | CASPIN | WEPIN | LEPIN
  ACTIVE_DEN =	DENPIN | CEPIN | CASPIN | WEPIN | LEPIN
  READ	     =	CEPIN | RASPIN | WEPIN | LEPIN
  READ_DEN   =	DENPIN | CEPIN | RASPIN | WEPIN | LEPIN
  WRITE	     =	CEPIN | RASPIN | LEPIN
  WRITE_LE   =	CEPIN | RASPIN
  WRITE_LE_DQM=	CEPIN | RASPIN | DQMPIN
  BURST_TERM =	CEPIN | RASPIN | CASPIN | LEPIN
  BURST_TERM_DEN = DENPIN | CEPIN | RASPIN | CASPIN | LEPIN
  PRECHARGE  =	CEPIN | CASPIN | LEPIN
  PRECHARGE_DEN= DENPIN | CEPIN | CASPIN | LEPIN
  AUTO_REFR  =	CEPIN | WEPIN | LEPIN
  LOADREG    =	CEPIN | LEPIN

VAR
  long dott_queue [Q_SIZE+1]
  long dott_qp
  long span_queue [Q_SIZE+1]
  long span_qp
  long char_queue [Q_SIZE+1]
  long char_qp
  long col_queue

  long cog
  long args [5]


PUB  Start (offsaddr)
  init_queues

{ Launch SDRAM cog }
  args[0] := @dott_queue
  args[1] := @span_queue
  args[2] := @char_queue
  args[3] := @col_queue
  args[4] := offsaddr      ' and the address in SDRAM of first scanline

  cog := 1 + cognew (@entry, @args) 'Launch new cog
  return cog

PUB  Stop
  if cog <> 0
    cogstop (cog-1)
    cog := 0

{ queue handling code }
PUB init_queues | i
  dott_qp := 0
  span_qp := 0
  char_qp := 0
  repeat i from 0 to Q_SIZE
    dott_queue [i] := 0
    span_queue [i] := 0
    char_queue [i] := 0
  col_queue := 0

PUB  drain_queues | i
  repeat i from 0 to Q_SIZE-1
    repeat until dott_queue [i] == 0
    repeat until span_queue [i] == 0
    repeat until char_queue [i] == 0
  repeat until col_queue == 0


PUB  set_colours (fore, back)
  drain_queues
  'repeat until col_queue == 0        ' here we wait until its been acknoledged as the common case.
  col_queue := $FFFF0000 | ((back & $FF) << 8) | (fore & $FF)
  repeat until col_queue == 0        ' here we wait until its been acknoledged as the common case.

PUB  draw_span (x, y, wid)
  repeat until span_queue [span_qp] == 0   ' await free slot
  span_queue [span_qp] := (x << 21) | ((y & $3FF) << 11) | (wid & $7FF)
  span_qp ++
  if span_qp => Q_SIZE
    span_qp := 0

PUB  draw_vline (x, y, h)
  repeat until dott_queue [dott_qp] == 0   ' await free slot
  dott_queue [dott_qp] := (x << 21) | ((y & $3FF) << 11) | (h & $3FF)
  dott_qp ++
  if dott_qp => Q_SIZE
    dott_qp := 0

PUB  draw_dot (x, y)
  draw_vline (x, y, 1)

PUB  draw_glyph (x, y, chr)
  repeat until char_queue [char_qp] == 0
  char_queue [char_qp] := (x << 21) | ((y & $3FF) << 11) | (chr & $FF)
  char_qp++
  if char_qp => Q_SIZE
    char_qp := 0


PUB  fill_rect (l, t, r, b) | i, j, w
  if l < r and t < b
    w := r-l+1
    repeat i from t to b
      draw_span (l, i, w)

PUB  draw_rect (l, t, r, b) | i, j, w, h
  if l < r and t < b
    w := r-l+1
    draw_span (l, t, w)
    draw_span (l, b, w)
    h := b-t-1
    if h == 0
      return
    repeat until h =< 64
      draw_vline (l, t+1, 64)
      draw_vline (r, t+1, 64)
      t += 64
      h -= 64
    draw_vline (l, t+1, h)
    draw_vline (r, t+1, h)
   

PUB  border_rect (l, t, r, b)
  fill_rect (l, t, r, b)
  set_colours (0,0)
  draw_rect (l, t, r, b)


PUB  do_circle (x, y, radius) | f, dx, dy, x1, y1   ' circle drawing code
  f := 1 - radius
  dx := 1
  dy := 1 - 2 * radius
  x1 := 0
  y1 := radius
  circle_slice (x-radius, x+radius, y)
  repeat
    if f => 0
      y1 --
      dy += 2
      f += dy
    x1 ++
    dx += 2
    f += dx
    circle_slice (x-x1, x+x1, y-y1)
    circle_slice (x-y1, x+y1, y-x1)
    circle_slice (x-y1, x+y1, y+x1)
    circle_slice (x-x1, x+x1, y+y1)
  until x1 => y1
'  circle_slice (x-radius, x+radius, y)  ' fix issue with first pixel?

PRI  circle_slice (xl, xh, y)
'  if xh>xl
    draw_span (xl, y, xh-xl+1)


DAT
		ORG	0

entry           mov     parm, PAR
		rdlong  dot_beg, parm
		add	parm, #4
		rdlong	spn_beg, parm
		add	parm, #4
		rdlong	chr_beg, parm
		add	parm, #4
		rdlong  col_q, parm
		add	parm, #4
		rdlong	scanbase, parm		' parm left pointing at scanbase

		mov	dot_qp, dot_beg
		mov	dot_end, dot_beg
		add	dot_end, queue_bytes

		mov	spn_qp, spn_beg
		mov	spn_end, spn_beg
		add	spn_end, queue_bytes

		mov	chr_qp, chr_beg
		mov	chr_end, chr_beg
		add	chr_end, queue_bytes


                mov     OUTA, #0
                or	DIRA, ADDR_MASK
                or      DIRA, CKEMASK           ' ensure CKE driven LOW
		or	DIRA, LCLKMASK          ' and LCLK

                mov	FRQA, CLOCK_FREQ
                mov     PHSA, PHASE
		movs	CTRA, #pin_SCLK
		movi	CTRA, #%0_00010_101	' PLL / 4, so 20MHz

                mov	FRQB, CLOCK_FREQ
                mov     PHSB, PHASE
		movs	CTRB, #pin_DCLK
		movi	CTRB, #%0_00010_101

		mov	VSCL, V4096
		mov	VCFG, vconfig_go
		WAITVID	idle_comms, #0	' force the video hardware to a known state
                or      DIRA, VIDMASK
                or      DIRA, SCLK_MASK		' now clocking in inactive state
                or      DIRA, DCLK_MASK

		mov	temp, CNT
		add	temp, hundred_us
		waitcnt	temp, delay
		call 	#init_sdram

top_of_frame
:loop           mov     count, Hheight
                rdlong  row_addr, scanbase
	        and     row_addr, ADDR_MASK
                mov     bank_addr, row_addr
                and     bank_addr, BANK_MASK
                mov     col_addr, bank_addr
                mov     VSCL, V50
:scanloop
                WAITVID refresh_comms, refresh_seq  ' synchronize with end of blanking
                call    #burst_read
                WAITVID refresh_comms, refresh_seq
                add     row_addr, #1
                and     row_addr, ADDR_MASK
                WAITVID refresh_comms, refresh_seq
                mov     bank_addr, row_addr
                and     bank_addr, BANK_MASK
                WAITVID refresh_comms, refresh_seq
                mov     col_addr, bank_addr
                djnz    count, #:scanloop

                WAITVID refresh_comms, refresh_seq
                mov     cycles_left, GAP_CYCLES
                mov     VSCL, Vwait
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
                jmp     #span_interp



refresh_comms   long    NONE | AUTO_REFR<<8
refresh_seq     long    %00_01_00_00_00_00_00_01_00_00

ba0_mask	long	1 << pin_BA0
ba1_mask	long	1 << pin_BA1
BANK_MASK       long    3 << pin_BA0
ADDR_MASK       long    $00007FFF

VIDMASK		long	$00FF0000
CKEMASK         long    1 << pin_CKE
LCLKMASK	long	1 << pin_LCLK
SCLK_MASK	long	1 << pin_SCLK
DCLK_MASK       long    1 << pin_DCLK
vconfig_stop	long    %0_00_1_0_0_000_00000_000000010_011111111  ' 2 bits/pixel VGA mode
vconfig_go	long    %0_01_1_0_0_000_00000_000000010_011111111


' single burst
burst_read      mov     VSCL, V800
                WAITVID read_comms1, read_seq1
                mov     OUTA, row_addr   ' ACTIVE
                mov     OUTA, col_addr    ' READ
                mov     VSCL, V50
                WAITVID read_comms2, read_seq2
		mov     OUTA, bank_addr     ' PRECHARGE bank
burst_read_ret  ret

{{
' switch-banks burst, allows more pixels per line
burst_read      mov     VSCL, V200
                WAITVID read_comms1, read_seq1
                mov     OUTA, row_addr   ' ACTIVE
                mov     OUTA, col_addr    ' READ
		add	row_addr, ba0_mask
		add	col_addr, ba0_mask
		
		mov	bcount, #3
:burst_loop
		WAITVID	read_commsx, read_seqx
		mov	OUTA, row_addr
		mov	OUTA, col_addr
		nop
		mov	OUTA, bank_addr
		add	row_addr, ba0_mask
		add	col_addr, ba0_mask
		add	bank_addr, ba0_mask
		djnz	bcount, #:burst_loop

                mov     VSCL, V50
                WAITVID read_comms2, read_seq2
		mov     OUTA, bank_addr     ' PRECHARGE bank
		sub	bank_addr, BANK_MASK
		and	row_addr, ADDR_MASK
		and	col_addr, ADDR_MASK
burst_read_ret  ret

bcount		long	0

V200            long    $1000 | 200
V300            long    $1000 | 300
V500            long    $1000 | 500
}}

{
' multi burst, allows horizontal scroll (upto a point)
burst_read      mov     VSCL, Vhscroll1
		add	col_addr, hscroll
                WAITVID read_comms1, read_seq1
                mov     OUTA, row_addr   ' ACTIVE
                mov     OUTA, col_addr    ' READ
		sub	col_addr, hscroll

		mov	VSCL, Vhscroll2
		WAITVID	read_commsx, read_seqxx
		nop
		mov	OUTA, col_addr

                mov     VSCL, V50
                WAITVID read_comms2, read_seq2
		mov     OUTA, bank_addr     ' PRECHARGE bank
burst_read_ret  ret

hscroll		long    scroll_amount
Vhscroll1	long	$1000 | (WIDTH - scroll_amount)
Vhscroll2	long	$1000 | scroll_amount
}


read_comms1     long    NONE | ACTIVE<<8 | READ<<16 | DEN<<24
read_seq1       long    %11_11_11_11_11_11_11_11_11_11_00_10_00_01_00_00

read_commsx	long	DEN | ACTIVE_DEN<<8 | READ_DEN<<16 | PRECHARGE_DEN<<24
read_seqx	long	%00_11_00_00_00_10_00_01_00_00
read_seqxx	long	%00_10_00_00_00_00


read_comms2     long    NONE | BURST_TERM_DEN<<8 | PRECHARGE<<16 | DEN<<24
read_seq2       long    %00_10_00_00_11_01_11_11_11_11





write_comms0    long    NONE | ACTIVE<<8 | WRITE_LE<<16 | BURST_TERM<<24
write_seq00     long    %11_10_00_01_00_00
write_seq0      long    %01_00_00 ' nop, active

write_comms1    long    NONE | LE<<8 | BURST_TERM<<16 | WRITE_LE<<24
write_seq1      long    %01_01_01_01_01_01_01_01_01_01_01_01_01_01_01_01  ' all LE
new_nar_seq	long	%10_01_01_01_01_01_01_01_01_01_01_01_01_01_01_01
new_wide_seq	long	%01_01_01_01_01_01_01_01_01_01_01_01_01_01_01_11


write_comms2    long    NONE | BURST_TERM<<8 | PRECHARGE<<16  ' $4A5A5E  ' NONE, BT, PRECHARGE
write_seq2      long    %00_10_00_01


char_setup      WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
                and     char, #$FF
         
		mov	OUTA, colour

		ror	char, #1  wc		' Prop font address decoding, C flag has odd/even
                shl     char, #7
                add     char, fontbase
                or      OUTA, LCLKMASK
                rdlong  patt, char		' first charline outside loop
		add	char, #4

                movs    VSCL, #24		' 16 pixels wide, 8 more for overheads
                WAITVID char_comms0, char_seq0   ' setup ACTIVE
                mov     OUTA, row_addr

                mov     count, #32		' charline count

	if_c    shr     patt, #1
                and     patt, H55555555
	  	or	patt, #2

:cloop
		movs	VSCL, #16
                mov     OUTA, col_addr
                WAITVID char_comms1, patt

                rdlong  patt, char

                movs    VSCL, #24
                WAITVID char_comms2, char_seq2  ' BT/PRECHARGE/NOP/ACTIVE
		add	char, #4
                mov     OUTA, bank_addr            ' precharge
                add     row_addr, #1            ' ready for next row
                mov     OUTA, row_addr          ' active

	if_c    shr     patt, #1		' select odd or even character
                and     patt, H55555555	        ' mask out other
	  	or	patt, #2                ' insert WRITE bit

                djnz    count, #:cloop

                mov     VSCL, Vwait
                WAITVID char_comms2, char_seq2x
                mov     OUTA, bank_addr     ' precharge
                sub     cycles_left, char_cycles
char_setup_ret  ret

char_cycles	long	24 + 32 * (16+24) + WAITCYCLES


char            long    0
patt            long    0
fontbase        long    $8000
H55555555       long    $55555555
D2              long    2 << 9
V9              long    $1000 | 9
V16             long    $1000 | 16
V24 		long	$1000 | 24

char_comms0     long    NONE | ACTIVE<<8  ' $4E5E      ' NONE, ACTIVE
char_seq0       long    %00_00_01_00_00

char_comms1     long    LE_DQM | LE<<8 | WRITE_LE_DQM<<16 | WRITE_LE<<24

char_comms2     long    NONE | BURST_TERM<<8 | PRECHARGE<<16 | ACTIVE<<24
char_seq2       long    %00_11_00_00_00_10_00_00_00_01
char_seq2x      long    %00_10_00_00_00


idle_comms  	long	IDLE
nop_comms       long    NONE

modeword	long	%00_0_00_010_0_111
precharge_all	long	$400		' A10 set

init_comms	long	NONE | PRECHARGE<<8 | LOADREG<<16    ' $424A5E ' NONE ,PRECHARGE,LOADREG
init_seq	long	%00_10_00_00_00_01_00_00


init_sdram     	mov	VSCL, V12
 		WAITVID init_comms, init_seq
		mov	OUTA, precharge_all
                nop
                mov	OUTA, modeword
                mov     VSCL, V50
		WAITVID nop_comms, #0

                mov     count, #1
		shl	count, #13
:loop
                WAITVID refresh_comms, refresh_seq  ' refresh every row
                djnz    count, #:loop

		mov	colour, #0
		mov	OUTA, colour
		nop
		or	OUTA, LCLKMASK
		andn	OUTA, LCLKMASK   ' initialize colour and latch

                WAITVID refresh_comms, refresh_seq  ' avoid waitvid timeout
init_sdram_ret	ret





{{ handle queued graphics operations }}

span_interp	cmp	cycles_left, MIN_CYCLES  wz,wc
	if_b	jmp	#wait_frame          ' run out of time, we're done


		' check the span queue first
:span_check	rdlong  temp, spn_qp  wz
	if_z	jmp	#:no_spans

		test	temp, H7F8  wz
 	if_nz	jmp	#:widespan
:narrowspan
		movs	VSCL, #24
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
		mov	row_addr, temp
		shr	row_addr, #11
		and	row_addr, H3FF
		mov	col_addr, temp
		shr	col_addr, #21
		mov	OUTA, colour
		mov	patt, new_nar_seq
		or	OUTA, LCLKMASK
		movs	VSCL, #26
		WAITVID write_comms0, write_seq0  ' #%10_00 ' nop, active
		mov	OUTA, row_addr

		and	temp, #7
		add	temp, temp
		sub	temp, #30
		neg	temp, temp
		shr	patt, temp
		or	patt, #%11

		mov	VSCL, #18
		sub	cycles_left, VSCL
		or	VSCL, H1000
		mov	OUTA, col_addr
		WAITVID write_comms1, patt
		mov	VSCL, Vwait
		mov	OUTA, #0
		WAITVID write_comms2, #%10_00
		sub	cycles_left, #WAITCYCLES+24+26
		jmp	#:donespan

:widespan
		movs	VSCL, #22
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
		mov	row_addr, temp
		shr	row_addr, #11
		and	row_addr, H3FF
		mov	col_addr, temp
		shr	col_addr, #21
		mov	OUTA, colour
		or	OUTA, LCLKMASK
		movs	VSCL, #18
		WAITVID write_comms0, write_seq0 ' nop, active
		mov	OUTA, row_addr
		and	temp, H7FF
'		add	temp, #8   ' patch
		mov	VSCL, temp
		sub	cycles_left, temp
		or	VSCL, H1000
		mov	OUTA, col_addr
		WAITVID write_comms1, new_wide_seq 
		mov	VSCL, Vwait
		mov	OUTA, #0
		WAITVID write_comms2, write_seq2
		sub	cycles_left, #WAITCYCLES+22+16


:donespan	wrlong	zero, spn_qp	     ' flag that we're done
		add	spn_qp, #4

		cmp	cycles_left, MIN_CYCLES  wz,wc
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
	if_ae	jmp	#:span_check
		jmp	#wait_frame          ' run out of time, we're done



:no_spans	cmp	spn_qp, spn_end  wz
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
	if_z	mov	spn_qp, spn_beg
	if_z	jmp	#:span_check


:dot_check	rdlong  temp, dot_qp  wz
	if_z	jmp	#:no_dots

		mov	row_addr, temp	 ' next 10 are y address
		shr	row_addr, #11
		and	row_addr, H3FF
		movs	VSCL, #28
		mov	OUTA, colour
                WAITVID refresh_comms, refresh_seq
		or	OUTA, LCLKMASK
                mov     col_addr, temp	 ' top 11 bits are X address
		shr	col_addr, #21
		and	col_addr, COLADDR_MASK
		mov	count, temp

		and	temp, H3FF
		shl	temp, #4
		sub	cycles_left, temp
		shr	temp, #2
		sub	cycles_left, temp

		shr	temp, #2
		movs	VSCL, #10
:loop
		WAITVID write_comms0, write_seq00
		mov	OUTA, row_addr
		mov	OUTA, col_addr
		add	row_addr, #1
		WAITVID write_comms2, #%10_00_00  ' NOP, PRECHARGE
		mov	OUTA, #0      
		djnz	temp, #:loop

		WAITVID nop_comms, #0
		mov	VSCL, Vwait
                WAITVID refresh_comms, refresh_seq
		sub	cycles_left, #WAITCYCLES+28+10+10

		wrlong	zero, dot_qp	     ' flag that we're done
		add	dot_qp, #4

		cmp	cycles_left, MIN_CYCLES  wz,wc
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
	if_ae	jmp	#:dot_check
		jmp	#wait_frame          ' run out of time, we're done



:no_dots	cmp	dot_qp, dot_end  wz
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
	if_z	mov	dot_qp, dot_beg
	if_z	jmp	#:dot_check





:char_check	rdlong	temp, chr_qp  wz
	if_z    jmp	#:no_chars

		call	#decode_address


                mov	char, temp
                and     char, #$FF		' font is 128 char one ATM
		call	#char_setup

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES



		wrlong	zero, chr_qp	     ' flag that we're done
		add	chr_qp, #4

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

		cmp	cycles_left, MIN_CYCLES  wz,wc
	if_ae	jmp	#:char_check
		jmp	#wait_frame          ' run out of time, we're done



:no_chars	cmp	chr_qp, chr_end  wz
                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES
	if_z	mov	chr_qp, chr_beg
	if_z	jmp	#:char_check

:col_check	rdlong	temp, col_q  wz

                WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

	if_z	jmp	#span_interp
		wrlong  zero, col_q

		mov	colour, temp
		and	colour, #$FF  ' currently limited to 8 bit colour

		WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

		jmp	#span_interp



wait_frame      cmps    cycles_left, #0  wz,wc
        if_be   jmp     #top_of_frame
                cmpsub  cycles_left, D2048  wc
        if_c    mov     VSCL, V2048
        if_nc   mov     VSCL, cycles_left
        if_nc   add     VSCL, WAITCYCLES
        if_nc   or      VSCL, V4096
        if_nc   mov     cycles_left, #0
                WAITVID refresh_comms, refresh_seq
                jmp     #wait_frame




decode_address  WAITVID refresh_comms, refresh_seq
                sub     cycles_left, #WAITCYCLES

                mov     col_addr, temp	 ' top 11 bits are X address
		shr	col_addr, #21
		and	col_addr, COLADDR_MASK
		mov	row_addr, temp	 ' next 10 are y address
		shr	row_addr, #11
		and	row_addr, H3FF
		mov	bank_addr, #0
decode_address_ret ret





hundred_us	long	8000
delay		long	800
queue_bytes	long	Q_BYTES

CLOCK_FREQ      long    $20000000  ' 5MHz * 16 / 4 = 40MHz PLL
PHASE           long    $00000000

' dataval		long	$000000FF ' 8 bit test pattern, may be 16bit later
writelen        long    250


Hheight         long    HEIGHT
COLADDR_MASK    long    $3FF
H3FF		long	$3FF
H7FF		long	$7FF

LOW_WORD        long    $0000FFFF
HIGH_WORD       long    $FFFF0000
D2048           long    $800

H1000		long	$1000
H7F8		long	$7F8
V4              long    $1000 | 4
V5              long    $1000 | 5
V12             long    $1000 | 12
V7              long    $1000 | 7
V11             long    $1000 | 11
V50             long    $1000 | 50
Vwait           long    $1000 | WAITCYCLES
V800            long    $1000 | 800
V2048           long    $1000 | 2048
V4096           long    $1000 | 4096

GAP_CYCLES      long    (WIDTH+HGAP)*VGAP
MIN_CYCLES      long    2000
zero            long    0


parm            res     1

dot_beg		res	1
dot_end		res	1
dot_qp		res	1
spn_beg		res	1
spn_end		res	1
spn_qp		res	1
chr_beg		res	1
chr_end		res	1
chr_qp		res	1
col_q		res	1

temp            res     1
count           res     1
colour          res     1
cycles_left     res     1

row_addr        res     1
col_addr        res     1
bank_addr       res     1
scanbase	res	1
lock		res	1

		FIT	$1F0

{{
////////////////////////////////////////////////////////////////////////////////////////////
//                                TERMS OF USE: MIT License
////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this 
// software and associated documentation files (the "Software"), to deal in the Software 
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////////////////
}}
